"""
test linear gaussian data gen
"""

from numpy.random import default_rng

from causalspyne.dag_gen import GenDAG

# from causalspyne.dag_manipulator import DAGManipulator
from causalspyne.gen_dag_2level import GenDAG2Level


def test_data_gen_linear_gaussian():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=30, degree=2, rng=default_rng(0))
    dag = simple_dag_gen.gen_dag()
    # DAGManipulator(dag)


def test_data_dag_subview():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=2, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete_test")



import numpy as np
from causallearn.search.ConstraintBased.FCI import fci
from causalspyne import gen_partially_observed

# visualization
from causallearn.utils.GraphUtils import GraphUtils


subview = gen_partially_observed(
    size_micro_node_dag=3,
    num_macro_nodes=2,
    degree=2,  # average vertex/node degree
    list_confounder2hide=[
        0.5,
        1.0,
    ],
    # choie of confounder to hide:
    # percentile or index of all toplogically sorted confounders
    num_sample=200,
    rng=np.random.default_rng(2),
    graphviz=True
)

# default parameters
g, edges = fci(subview.data)


print(g.graph)  # numpy array

"""
see (https://causal-learn.readthedocs.io/en/latest/search_methods_index/Constraint-based%20causal%20discovery%20methods/FCI.html#returns) for how to interpret FCI output

g.graph with rng(0)
array([[0, 0, 0, 0],
      [0, 0, 0, 0],
      [0, 0, 0, 2],
      [0, 0, 2, 0]])
here node4-5 connection type 2 edge means non-determined o-o type edge
g.graph with rng(1)
[[0 2 0 0 2]
[2 0 2 0 2]
[0 2 0 2 2]
[0 0 2 0 0]
[2 2 2 0 0]]

g.graph with rng(2)
1 means directed edge
[[0 2 0 0 0 0]
[1 0 1 0 0 0]
[0 2 0 2 2 0]
[0 0 2 0 0 0]
[0 0 1 0 0 1]
[0 0 0 0 2 0]]
"""
pdy = GraphUtils.to_pydot(g, labels=subview.node_names)
pdy.write_png("output_causallearn_fci.png")



import argparse
import configparser
# ConfigArgParse is a drop-in replacement for argparse that allows options to
# be set via config files and environment variables.


class DynamicConfig:
    def __init__(self, conf):
        for key, value in conf.items():
            setattr(self, key, value)


def parse_args_and_config():
    parser = argparse.ArgumentParser()
    # parser = configargparse.ArgumentParser(description='Your program description')
    parser.add_argument('-c', '--config', help='Config file path')
    parser.add_argument('--option1', help='Option 1 description')
    parser.add_argument('--option2', help='Option 2 description')
    args = parser.parse_args()

    config = configparser.ConfigParser()
    if args.config:
        config.read(args.config)

    # Merge config file and command line arguments
    options = {}
    if 'DEFAULT' in config:
        options.update(config['DEFAULT'])
    options.update({k: v for k, v in vars(args).items() if v is not None})

    return options
def load_config_from_file(file_path):
    config = configparser.ConfigParser()
    config.read(file_path)
    return config

def parse(args):
    if args.config:
        config = load_config_from_file(args.config)
        db_host = config.get('database', 'host', fallback=args.db_host)
        db_port = config.getint('database', 'port', fallback=args.db_port)
        db_user = config.get('database', 'user', fallback=args.db_user)
        db_password = config.get('database', 'password', fallback=args.db_password)
        db_name = config.get('database', 'dbname', fallback=args.db_name)
        server_host = config.get('server', 'host', fallback=args.server_host)
        server_port = config.getint('server', 'port', fallback=args.server_port)
    else:
        db_host = args.db_host
        db_port = args.db_port
        db_user = args.db_user
        db_password = args.db_password
        db_name = args.db_name
        server_host = args.server_host
        server_port = args.server_port

def test_args():
    options = parse_args_and_config()
    print(options)





"""Generate ground truth pair (D, A_S), where D is a DAG and A is the
optimal abstraction of D with respect to score S (e.g. BIC, AIC score).

Given D, construct the poset of graphical abstractions and perform
exhaustive search for the optimal abstraction. This will be a useful
ground truth for evaluating a greedy algorithm that avoids learning D
and directly learns A_S via greedy search from the top of the lattice.

We can start in the observational setting at first and then move to
the interventional setting later.
"""

from typing import Generator, List


def partitions(n: int) -> Generator[List[List[int]], None, None]:
    """Enumerate all partitions of n labelled objects.

    The number of partitions is given by the Bell number, B_n. See
    https://en.wikipedia.org/wiki/Partition_of_a_set#Counting_partitions

    Make sure to memoize this first if we plan to call it many times!
    """
    if n == 1:
        yield [[1]]
    else:
        for partition in partitions(n - 1):
            yield partition + [[n]]
            for idx, part in enumerate(partition):
                prepart = partition[:idx]
                postpart = partition[idx + 1:]
                yield prepart + [part + [n]] + postpart


def abstractions(partial_order) -> Generator[List, None, None]:
    """Enumerate partitions that are consistent with the given partial
    order."""
    yield None



"""
test 2 level DAG generation
"""

from numpy.random import default_rng

from causalspyne.gen_dag_2level import GenDAG2Level
from causalspyne.dag_gen import GenDAG
from causalspyne.is_dag import is_dag
from causalspyne.draw_dags import draw_dags_nx


def test_gen_dag_2level():
    """
    test 2 level generation of DAGs
    """

    dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    gen = GenDAG2Level(dag_generator=dag_gen, num_macro_nodes=3, rng=default_rng(0))
    dag = gen.run()
    dag
    print(dag)
    is_dag(dag.mat_adjacency)
    draw_dags_nx(dag.mat_adjacency, dict_ind2name=dag.gen_dict_ind2node_na())



"""
generate DAG and its marginal DAG
"""

from datetime import datetime

try:
    from contextlib import chdir
except Exception:
    from causalspyne.py3_9_10_compatibility import chdir

from pathlib import Path

import matplotlib.pyplot as plt
from numpy.random import default_rng

from causalspyne.gen_dag_2level import GenDAG2Level
from causalspyne.dag_gen import GenDAG
from causalspyne.dag_viewer import DAGView
from causalspyne.dag2ancestral import DAG2Ancestral

from causalspyne.draw_dags import draw_dags_nx


def gen_partially_observed(
    degree=2,
    list_confounder2hide=[0.5, 0.9],
    size_micro_node_dag=4,
    num_macro_nodes=4,
    num_sample=200,
    output_dir="output/",
    rng=default_rng(),
    dft_noise="Gaussian",
    graphviz=False,
    plot=True,
):
    """
    sole function as user interface
    """
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_")
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    simple_dag_gen = GenDAG(num_nodes=size_micro_node_dag,
                            degree=degree, rng=rng)

    # num_macro_nodes will overwrite behavior
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen,
        num_macro_nodes=num_macro_nodes, rng=rng
    )
    dag = dag_gen.run()
    dag.to_binary_csv(benchpress=False,
                      name=output_dir + f"ground_truth_dag_{timestamp}d.csv")

    subview = DAGView(dag=dag, rng=rng, dft_noise=dft_noise)
    return re_hide(subview, dag, num_sample, list_confounder2hide, output_dir,
                   graphviz, timestamp, plot=True)


def ordered_ind_col2global_ind(inds_cols, subview_global_inds):
    """
    given a predicted causal order in the form of column indices, transform it
    into global index of ground truth DAG
    """
    list_global_inds = [subview_global_inds[ind_col] for ind_col in inds_cols]
    return list_global_inds


def re_hide(subview, dag, num_sample, list_confounder2hide, output_dir,
            graphviz, timestamp, plot=True):
    subview.run(
        num_samples=num_sample, confound=True,
        list_nodes2hide=list_confounder2hide
    )
    with chdir(output_dir):
        subview.to_csv()
    str_node2hide = subview.str_node2hide

    dag2ancestral = DAG2Ancestral(dag.mat_adjacency)
    list_confounder2hide_global_ind = subview.list_global_inds_nodes2hide
    pred_ancestral_graph_mat = dag2ancestral.run(
        list_confounder2hide_global_ind)

    if plot:
        fig, (ax1, ax2, ax3) = plt.subplots(1, 3)
        mtitle = "hide_" + str_node2hide
        fig.suptitle(mtitle)  # super-title

        # ax1
        dag.visualize(title="DAG", ax=ax1, graphviz=graphviz)
        ax1.set_title("DAG")

        # ax2
        draw_dags_nx(
            pred_ancestral_graph_mat,
            dict_ind2name={
                i: name for i, name in enumerate(sorted(subview.node_names))
            },
            title="ancestral",
            ax=ax2,
            graphviz=graphviz,
        )
        ax2.set_title("ancestral")
        # ax3
        subview.visualize(
            title="subDAG", ax=ax3, graphviz=graphviz
        )
        ax3.set_title("subDAG")

    with chdir(output_dir):
        subview.to_csv()
        if plot:
            fig.savefig(f"graph_compare_{timestamp}dags.pdf", format="pdf")
            fig.savefig(f"graph_compare_{timestamp}dags.svg", format="svg")
        with open("hidden_nodes.csv", "w") as outfile:
            outfile.write(
                ",".join(str(node) for node in
                         subview._list_global_inds_unobserved)
            )
    return subview



import numpy as np


class EdgeModelLinear():
    def __init__(self, dag):
        self._dag = dag

    def run(self, ind_node, data_realization_parents):
        """
        data_realization_parents = data[:, list_parents_inds]
        """
        # Linear combination of parent nodes + Gaussian noise
        weights = self._dag.get_weights_from_list_parents(ind_node)
        bias = np.dot(data_realization_parents, weights)
        return bias


class EdgeModelIfElse():
    def __init__(self, dag):
        self._dag = dag

    def run(self, ind_node, data_realization_parents):
        """
        data_realization_parents = data[:, list_parents_inds]
        """
        # Linear combination of parent nodes + Gaussian noise
        weights = self._dag.get_weights_from_list_parents(ind_node)
        weights_mod = np.where(data_realization_parents == 0,
                               -weights, weights)
        bias = np.dot(data_realization_parents, weights_mod)
        return bias



import numpy as np
from causalspyne.utils_causallearn_g2ancestral import get_causalearn_order
from causalspyne import gen_partially_observed, ordered_ind_col2global_ind
from causalspyne.ancestral_acc import ancestral_acc

from causallearn.search.HiddenCausal.GIN.GIN import GIN
from causallearn.graph.NodeType import NodeType

# visualization
from causallearn.utils.GraphUtils import GraphUtils
# Visualization using pydot
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import io


subview = gen_partially_observed(
    size_micro_node_dag=3,
    num_macro_nodes=2,
    degree=2,  # average vertex/node degree
    list_confounder2hide=[
        0.5,
        1.0,
    ],  # choie of confounder to hide: percentile or index of all toplogically sorted confounders
    num_sample=200,
    rng=np.random.default_rng(1),
    graphviz=True
)
print(f"subview global inds: {subview.col_inds}")
G, K = GIN(subview.data)

pred_obs_order, pred_latent_order = get_causalearn_order(G, subview.node_names)
print(f"predicted observable order:{pred_obs_order}")

ind_cols = [int(name.removeprefix("X")) for name in pred_obs_order]
# pred_obs_order_aligned = ordered_ind_col2global_ind(ind_cols, subview_global_inds)
acc = ancestral_acc(subview.dag, pred_order=ind_cols)

print(f"acc: {acc}")

print(f"latent cluster order: {K}, type: {type(K)}, len: {len(K)}")

print(G.graph)  # numpy array of PAG,
# see https://causal-learn.readthedocs.io/en/latest/search_methods_index/Constraint-based%20causal%20discovery%20methods/FCI.html#usage

# FIXME: directly providing pred_obs_order to "to_pydot" cause inconsistency of GraphUtils
pyd = GraphUtils.to_pydot(G)
pyd.write_png("output_causallearn_gin.png")

# for showing image
# tmp_png = pyd.create_png(f="png")
# fp = io.BytesIO(tmp_png)
# img = mpimg.imread(fp, format='png')
# plt.axis('off')
# plt.imshow(img)
# plt.show()



from numpy.random import default_rng

from causalspyne.erdo_renyi_plp import Erdos_Renyi_PLP
from causalspyne.is_dag import is_dag


def test_erdos_renyi():
    mat1 = Erdos_Renyi_PLP(default_rng(0))(num_nodes=3, degree=2)
    is_dag(mat1)



"""
class method for DAG operations and check
"""

import numpy as np
from numpy.random import default_rng
import pandas as pd

from causalspyne.is_dag import is_dag
from causalspyne.utils_topological_sort import topological_sort
from causalspyne.draw_dags import draw_dags_nx


def add_prefix(string, prefix="", separator="u"):
    """
    Adds a prefix to a string.
    If the prefix is empty, returns the original string.

    Args:
    string (str): The original string.
    prefix (str, optional): The prefix to add. Defaults to an empty string.

    Returns:
    str: The string with the prefix added.
    """
    if not prefix:
        return string
    return separator.join([prefix, string])


class MatDAG:
    """
    DAG represented as a mat_adjacency
    """

    def __init__(
        self,
        mat_adjacency,
        name_prefix="",
        separator="_",
        list_node_names=None,
        parent_list_node_names=None,
        rng=default_rng(0),
    ):
        """ """
        self._obj_gen_weight = None
        self.separator = separator
        self.name_prefix = name_prefix
        self.mat_adjacency = mat_adjacency
        self._list_node_names = list_node_names
        self._list_confounder = None
        self._dict_node_names2ind = {}
        self._parent_list_node_names = parent_list_node_names
        self._init_map()
        self._list_ind_nodes_sorted = None
        self.rng = rng

    def _init_map(self):
        if self._list_node_names is not None:
            self._dict_node_names2ind = {
                name: i for (i, name) in enumerate(self._list_node_names)
            }

    @property
    def list_confounder(self):
        """
        return list of confounders
        """
        nonzero_counts = np.count_nonzero(self.mat_adjacency, axis=0)
        columns_with_more_than_one = np.where(nonzero_counts > 1)[0]
        return list(columns_with_more_than_one)

    def gen_dict_ind2node_na(self, hierarch_na=False):
        """
        utility function to have {1:"node_name"} dictionary for plotting
        hierarch_na: if use hierarchical name maco-node-micro-node format
        """
        if hierarch_na:
            mdict = {i: name for (i, name) in enumerate(self.list_node_names)}
        elif self._parent_list_node_names is not None:
            mdict = {
                i: str(self._parent_list_node_names.index(name))
                for (i, name) in enumerate(self.list_node_names)
            }
        else:
            # FIXME: if self.list_node_names is list of string,
            # then it will be ignored, mdict will be {0:integer} style
            # always
            mdict = {
                i: str(self.list_node_names.index(name))
                for (i, name) in enumerate(self.list_node_names)
            }
        return mdict

    def check(self):
        """
        check if the matrix represent a DAG
        """
        if not is_dag(self.mat_adjacency):
            raise RuntimeError("not a DAG")
        binary_adj_mat = (self.mat_adjacency != 0).astype(int)
        if not is_dag(binary_adj_mat):
            raise RuntimeError("not a DAG")

    @property
    def num_nodes(self):
        """
        number of nodes in DAG
        """
        return self.mat_adjacency.shape[0]

    def gen_node_names(self):
        """
        get list of node names
        """
        self._list_node_names = [
            add_prefix(string="v" + str(i), prefix=self.name_prefix)
            for i in range(self.num_nodes)
        ]
        self._init_map()

    def gen_node_names_stacked(self, dict_macro_node2dag):
        self._list_node_names = []
        for key, dag in dict_macro_node2dag.items():
            self._list_node_names.extend(dag.list_node_names)
        self._init_map()

    def get_node_ind(self, node_name):
        """
        get the matrix index of the node name
        """
        return self._dict_node_names2ind[node_name]

    @property
    def list_node_names(self):
        """
        get the node names in a linear list
        """
        if self._list_node_names is None:
            self.gen_node_names()
        return self._list_node_names

    @property
    def list_arcs(self):
        """
        return the list of edges
        """
        list_i_j = list(zip(*self.mat_adjacency.nonzero()))
        list_arcs = [
            (self.list_node_names[tuple(ij)[0]],
             self.list_node_names[tuple(ij)[1]])
            for ij in list_i_j
        ]
        return list_arcs

    def sample_node(self):
        """
        randomly chose a node
        """
        ind = self.rng.integers(0, self.mat_adjacency.shape[0])
        name = self.list_node_names[ind]
        return self.name_prefix + name, ind

    def add_arc_ind(self, ind_tail, ind_head, weight=None):
        """
        add arc via index of tail and head
        """
        node_tail = self.list_node_names[ind_tail]
        node_head = self.list_node_names[ind_head]
        self.add_arc(node_tail, node_head, weight)

    def add_arc(self, node_tail, node_head, weight=None):
        """
        add edge to adjacency matrix
        """
        ind_tail = self._dict_node_names2ind[node_tail]
        ind_head = self._dict_node_names2ind[node_head]
        if weight is None:
            self.mat_adjacency[ind_tail, ind_head] = 1
        else:
            self.mat_adjacency[ind_tail, ind_head] = weight

    def to_binary_csv(self, benchpress=True, name="adj.csv"):
        """
        adjacency matrix to csv format
        """
        binary_adj_mat = (self.mat_adjacency != 0).astype(int)
        if benchpress:
            binary_adj_mat = np.transpose(binary_adj_mat)
        df = pd.DataFrame(binary_adj_mat, columns=self.list_node_names)
        df.to_csv(name, index=False)

    def topological_sort(self):
        """
        topological sort DAG into list of node index
        """
        binary_adj_mat = (self.mat_adjacency != 0).astype(int)
        self._list_ind_nodes_sorted = topological_sort(binary_adj_mat)
        return self._list_ind_nodes_sorted

    @property
    def list_ind_nodes_sorted(self):
        """
        get global node index topologically sorted
        """
        if self._list_ind_nodes_sorted is None:
            self.topological_sort()
        return self._list_ind_nodes_sorted

    def get_list_parents_inds(self, ind_node):
        """
        get list of parents nodes
        """
        # np.nonzero(x)
        # returns (array([0, 1, 2, 2]), array([0, 1, 0, 1]))
        # assume lower triangular matrix as adjacency matrix
        # matrix[i, j]=1 indicate arrow j->i
        submatrix = self.mat_adjacency[ind_node, :]
        vector = submatrix.flatten()
        list_inds = np.nonzero(vector)[0].tolist()
        return list_inds

    def get_weights_from_list_parents(self, ind_node):
        """
        get incoming edge weights
        """
        list_parents_inds = self.get_list_parents_inds(ind_node)
        sub_matrix = self.mat_adjacency[ind_node, list_parents_inds]
        return sub_matrix

    def __str__(self):
        return str(self.mat_adjacency)

    def __repr__(self):
        return str(self.mat_adjacency)

    def subgraph(self, list_ind_unobserved):
        """
        subset adjacency matrix by deleting unobserved variables
        """
        # delete first axis
        temp_mat_row = np.delete(
            self.mat_adjacency, list_ind_unobserved, axis=0)

        # delete second axis
        mat_adj_subgraph = np.delete(
            temp_mat_row, list_ind_unobserved, axis=1)

        # filter out subgraph node names
        list_node_names_subgraph = [
            x
            for i, x in enumerate(self.list_node_names)
            if i not in list_ind_unobserved
        ]

        subdag = MatDAG(
            mat_adj_subgraph,
            list_node_names=list_node_names_subgraph,
            parent_list_node_names=self.list_node_names,
            rng=self.rng,
        )
        return subdag

    def visualize(self, title="dag", hierarch_na=False, ax=None,
                  graphviz=False):
        """
        draw dag using networkx
        """
        draw_dags_nx(
            self.mat_adjacency,
            dict_ind2name=self.gen_dict_ind2node_na(hierarch_na),
            title=title,
            ax=ax,
            graphviz=graphviz
        )

    @property
    def list_top_names(self):
        """
        return list of node names in toplogical order
        """
        list_top_names = [self.list_node_names[i] for i in
                          self.list_ind_nodes_sorted]
        return list_top_names

    def global_arbitrary_ind2topind(self, ind_global_arbitrary):
        ind_top = self.list_ind_nodes_sorted.index(ind_global_arbitrary)
        return ind_top

    def top_ind2global_arbitrary(self, ind_top):
        ind_global_arbitrary = self.list_ind_nodes_sorted[ind_top]
        return ind_global_arbitrary

    def get_top_last(self):
        ind_global_arbitrary = self.list_ind_nodes_sorted[-1]
        return ind_global_arbitrary

    def climb(self, ind_arbitrary):
        ind_top = self.global_arbitrary_ind2topind(ind_arbitrary)
        if ind_top - 1 >= 0:
            ind_arbi = self.top_ind2global_arbitrary(ind_top - 1)
            return ind_arbi
        return None

    @property
    def num_confounder(self):
        return len(self.list_confounder)

    @property
    def list_top_order_sorted_confounder(self):
        list_top_oder_confounder = [
            self.list_ind_nodes_sorted.index(confounder)
            for confounder in self.list_confounder
        ]
        list_top_oder_confounder_sorted = sorted(list_top_oder_confounder)
        return list_top_oder_confounder_sorted



import numpy as np


def transitive_closure(adj_matrix):
    """
    Floyd-Warshall algorithm, note that this algorithm is agnostic w.r.t.
    the edge convention: (i,j) implies either i<-j or i->j is the same
    for the algorithm
    """
    num_vertex = len(adj_matrix)
    closure = np.array(adj_matrix, dtype=bool)

    for k in range(num_vertex):  # intermiediate connection
        for i in range(num_vertex):
            for j in range(num_vertex):
                # as long as one of the three cases result in connection, then
                # i, j should be connected
                closure[i][j] = closure[i][j] or \
                    (closure[i][k] and closure[k][j])
                # i-k, k-j implies i-j
                # here the mathematical relation i-j or i~j, can mean
                # both i->j or j<-i
    return closure


def ancestor_matrix(adj_matrix):
    num_vertex = len(adj_matrix)
    closure = transitive_closure(adj_matrix)
    ancestor = np.zeros((num_vertex, num_vertex), dtype=bool)

    for i in range(num_vertex):
        for j in range(num_vertex):
            if i != j:
                ancestor[i][j] = closure[i][j]
    return ancestor



"""
"""
import numpy as np
from causalspyne.implicit_omega_condition_on_w import gen_joint_w_omega, get_max_degree
from causalspyne.implicit_omega import get_extreme_eigenvalue


def cov_to_corr(covariance_matrix):
    # Calculate the standard deviations
    std_devs = np.sqrt(np.diag(covariance_matrix))

    # Create a diagonal matrix of inverse standard deviations
    inv_std_devs = np.diag(1 / std_devs)

    # Calculate the correlation matrix
    correlation_matrix = inv_std_devs @ covariance_matrix @ inv_std_devs

    return correlation_matrix


def congruent(mat_op, mat):
    """
    congruent transform is defined to be (mat_op)^T mat (mat_op)
    """
    first2 = np.matmul(mat_op.T, mat)
    mat_rst = np.matmul(first2, mat_op)
    return mat_rst


def gen_sigma_y(max_w=0.5):
    """
    generate Sigma matrix according to $${(I-W)}^{-1}\Omega{(I-W)}^{-1}$$
    """
    mat_w_binary, mat_omega = gen_joint_w_omega()
    mat_weight = np.random.uniform(
        low=-max_w, high=max_w, size=mat_w_binary.shape)

    mat_w = mat_w_binary * mat_weight
    print(f"adjacency weighted: {mat_w}")

    dmax = get_max_degree(mat_w)
    kernel = np.identity(mat_w.shape[0]) - mat_w
    inv_id_minus_w = np.linalg.inv(kernel)
    mat_sigma = congruent(inv_id_minus_w.T, mat_omega)
    max_omega = np.max(mat_omega)
    return mat_sigma, dmax, max_omega


def gen_spectrum():
    mat_sigma, dmax, max_omega = gen_sigma_y()
    print(f"sigma: {mat_sigma}")
    inv_sigma = np.linalg.inv(mat_sigma)
    eigv_max, _ = get_extreme_eigenvalue(inv_sigma)
    print(f"max eigen value: {eigv_max}")
    print(f"max degree: {dmax}, max_omega: {max_omega}")
    mat_precision = cov_to_corr(inv_sigma)
    eigv_max, _ = get_extreme_eigenvalue(mat_precision)
    print(f"max eigen value: {eigv_max}")
    return eigv_max


if __name__ == "__main__":
    gen_spectrum()



import numpy as np

from causalspyne.draw_dags import draw_dags_nx


def test_draw_dag_nx():
    # Create an example adjacency matrix for a DAG
    adj_matrix = np.array([[0, 1, 1, 0], [0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]])
    draw_dags_nx(adj_matrix=adj_matrix, show=False)



import numpy as np
from causalspyne import gen_partially_observed, ordered_ind_col2global_ind
from causalspyne.main import ordered_ind_col2global_ind
from causalspyne.ancestral_acc import ancestral_acc

import graphviz
import lingam
from lingam.utils import print_causal_directions, print_dagc, make_dot


subview = gen_partially_observed(
    size_micro_node_dag=3,
    num_macro_nodes=2,
    degree=2,  # average vertex/node degree
    list_confounder2hide=[
        0.5,
        1.0,
    ],  # choie of confounder to hide: percentile or index of all toplogically sorted confounders
    num_sample=200,
    rng=np.random.default_rng(1),
    graphviz=True
)


model = lingam.BottomUpParceLiNGAM()
model.fit(subview.data)

print(f"predicted observable order:{model.causal_order_}")

nested_list = model.causal_order_

print(f"causal order {nested_list}")

flat_list = [item for sublist in nested_list for item in (sublist if isinstance(sublist, list) else [sublist])]


pred_obs_order = ordered_ind_col2global_ind(inds_cols=flat_list,
                                            subview_global_inds=subview.col_inds)


acc = ancestral_acc(subview.dag, pred_order=pred_obs_order)

print(f"ancestral acc:{acc}")


make_dot(model.adjacency_matrix_)



"""
create different views for the same DAG by hiding some variables
"""

import warnings

import numpy as np
from numpy.random import default_rng
import pandas as pd

from causalspyne.data_gen import DataGen
from causalspyne.dag_interface import MatDAG


def process_list2hide(list_ind_or_percentage, total_num):
    """
    list_ind_or_percentage can either be a list or a scalar float
    """
    if len(list_ind_or_percentage) > total_num:
        raise RuntimeError(
            f"there are {total_num} confounders to hide, less \
                           than the length of {list_ind_or_percentage}"
        )

    list_ind = [
        min(int(ele * total_num), total_num - 1)
        if isinstance(ele, float) else ele
        for ele in list_ind_or_percentage
    ]

    list_ind = list(set(list_ind))

    if max(list_ind) > total_num:
        raise RuntimeError(
            f"max value in {list_ind_or_percentage} is bigger \
                           than total number of variables {total_num} to hide"
        )

    return list_ind


class DAGView:
    """
    with ground truth DAG intact, only show subgraph
    """

    def __init__(self, dag, dft_noise, rng=default_rng(0)):
        self._dag = dag
        # there is no need to use a full DAG to represent subdag
        # since sub-dag is not responsible for data generation
        self._sub_dag = None
        self._data_arr = None
        self._subset_data_arr = None
        self._list_global_inds_unobserved = None
        self._list_global_inds_observed = None
        self.data_gen = DataGen(dag=self._dag,
                                dft_noise=dft_noise,
                                rng=rng)
        self._list_nodes2hide = None
        self._success = False

    @property
    def dag(self):
        return self._dag

    def run(self, num_samples, list_nodes2hide=None, confound=False):
        """
        generate subgraph adjcency matrix and corresponding data
        """
        self._data_arr = self.data_gen.gen(num_samples)
        if list_nodes2hide is None:
            list_nodes2hide = [0]
        if confound:
            self.hide_confounder(list_nodes2hide)
        else:
            self.hide_top_order(list_nodes2hide)

    def hide_confounder(self, list_toporder_confounder2hide_input):
        """
        given a list of index, hide the confounder according to the toplogical
        order provided by the input index list_toporder_confounder2hide
        then call self.hide
        """
        if not self._dag.list_confounder:
            raise RuntimeError(
                f"there are no confounders in the graph {self._dag} \
                          !"
            )

        list_toporder_confounder2hide = process_list2hide(
            list_toporder_confounder2hide_input, len(self._dag.list_confounder)
        )

        list_ind_confounder_sorted = self._dag.list_top_order_sorted_confounder

        list_toporder_confounder_sub = [
            list_ind_confounder_sorted[i]
            for i in list_toporder_confounder2hide
        ]

        self.hide_top_order(list_toporder_confounder_sub)
        return True

    def hide_top_order(self, list_toporder_unobserved):
        """
        hide variables according to a list of global index of topological sort
        """
        list_toporder_unobserved = process_list2hide(
            list_toporder_unobserved, self._dag.num_nodes
        )

        # subset list
        self._list_nodes2hide = [
            self._dag.list_top_names[i] for i in list_toporder_unobserved
        ]

        # FIXME: change to logger
        print("nodes to hide " + str(self._list_nodes2hide))

        self._list_global_inds_unobserved = [
            self._dag.list_ind_nodes_sorted[ind_top_order]
            for ind_top_order in list_toporder_unobserved
        ]

        self._sub_dag = self._dag.subgraph(self._list_global_inds_unobserved)
        self._subset_data_arr = np.delete(
            self._data_arr, self._list_global_inds_unobserved, axis=1
        )
        self._success = True

    @property
    def data(self):
        """
        return data in numpy array format
        """
        return self._subset_data_arr

    @property
    def mat_adj(self):
        """
        return adj matrix
        """
        return self._sub_dag.mat_adjacency

    def check_if_subview_done(self):
        """
        check if DAG marginalizatino successfull or not
        """
        if not self._success:
            warnings.warn("no subview of DAG available, exit now!")
            return

    @property
    def node_names(self):
        # filter out observed variable
        # _node_names = [name for (i, name) in
        #               enumerate(self._dag.list_node_names)
        #               if i not in self._list_global_inds_unobserved]
        _node_names = self._sub_dag.list_node_names
        _node_names_ind = ["X" + str(self._sub_dag._parent_list_node_names.index(name)) for
        name in _node_names]
        return _node_names_ind

    def to_csv(self, title="data_subdag.csv"):
        """
        sub dataframe to  csv
        """
        self.check_if_subview_done()

        # FIXME: ensure self.node_names are consistent with self.data
        df = pd.DataFrame(self.data, columns=self.node_names)
        df.to_csv(title[:-4] + "_" + self.str_node2hide + title[-4:],
                  index=False)

        subdag = MatDAG(self.mat_adj)
        subdag.to_binary_csv()

    @property
    def list_global_inds_nodes2hide(self):
        return self._list_global_inds_unobserved

    @property
    def col_inds(self):
        subview_global_inds = \
            [self.dag._dict_node_names2ind[name]
             for name in self.dag.list_node_names
             if name not in self.str_node2hide]
        return subview_global_inds

    @property
    def list_global_inds_observed(self):
        if self._list_global_inds_observed is None:
            if self._list_global_inds_unobserved is None:
                raise RuntimeError(
                    "global inds for unobserved not initialized yet!")
            self._list_global_inds_observed = \
                [item for item in self._dag.list_ind_nodes_sorted
                 if item not in self._list_global_inds_unobserved]
        return self._list_global_inds_observed

    @property
    def str_node2hide(self):
        """
        string representation of nodes to hide(marginalize)
        """
        self.check_if_subview_done()
        if self._list_nodes2hide is None:
            raise RuntimeError("self._list_node2hide is None!")
        _str_node2hide = "_".join(map(str, self._list_nodes2hide))

        _str_node2hide_ind = "_".join(map(str, self._list_global_inds_unobserved))

        return "_ind_".join([_str_node2hide, _str_node2hide_ind])

    def visualize(self, **kwargs):
        """
        plot DAG
        """
        if not self._success:
            warnings.warn("no subview of DAG available")
            return
        self._sub_dag.visualize(**kwargs)



"""
This module is used to generate the implicit omega matrix.

To generate a symmetric matrix,
    - A+A^T
    - np.tril(A) + np.tril(A, -1).T
    np.tril(A, -1) returns a copy of the input array A with elements above the
    first sub-diagonal zeroed out
"""
import numpy as np


def is_diagonal(matrix):
    return np.count_nonzero(matrix - np.diag(np.diag(matrix))) == 0


def is_diag_dom(mat):
    for i_row, row in enumerate(mat):
        energy_diag = np.abs(mat[i_row, i_row])
        energy_row_off_diag = np.sum(np.abs(row)) - energy_diag
        flag = energy_diag <= energy_row_off_diag
        if flag:
            return False
    return True


def gen_omega(n_ob_v, max_omega, delta=1.0, p_sparse=0.1):
    """
    :n_ob_v: number of observed variables
    :max_omega: maximum value of omega
    :returns: None
    """
    assert max_omega > 0
    # Generate Bernoulli random numbers
    mat_bernoulli = np.random.binomial(
        n=1, p=p_sparse, size=(n_ob_v, n_ob_v))
    mat_omega = np.random.uniform(
        low=0.1, high=max_omega, size=(n_ob_v, n_ob_v))
    mat_sparse_omega = mat_bernoulli * mat_omega
    mat_sym = 0.5 * (mat_sparse_omega + mat_sparse_omega.T)

    if is_diagonal(mat_sym):
        rng = np.random.default_rng()
        sampled_row_ind = rng.integers(low=0, high=n_ob_v - 1, size=1)
        off_diag_v = mat_omega[sampled_row_ind, sampled_row_ind + 1]
        mat_sym[sampled_row_ind, sampled_row_ind + 1] = off_diag_v
        mat_sym[sampled_row_ind + 1, sampled_row_ind] = off_diag_v
    for i_row, row in enumerate(mat_sym):
        old_diag_v = np.abs(mat_sym[i_row, i_row])
        energy_row_off_diag = np.sum(np.abs(row)) - np.abs(old_diag_v)
        mat_sym[i_row, i_row] = delta + energy_row_off_diag
    assert is_diag_dom(mat_sym)
    return mat_sym, delta + max_omega


def get_extreme_eigenvalue(matrix):
    """
    return: max, min
    """
    eigenvalues = np.linalg.eigvals(matrix)
    return np.max(np.real(eigenvalues)), np.min(np.real(eigenvalues))


if __name__ == "__main__":
    max_omega = 0.7
    mat, rho = gen_omega(3, max_omega)
    print(mat)
    max_eig, min_eig = get_extreme_eigenvalue(mat)
    print(f"min, max eigenvalue {min_eig, max_eig}")
    print(rho, 2 * rho)



"""
concrete class to generate simple DAGs

"""
import numpy as np


class Erdos_Renyi_PLP:
    """
    uniformly (w.r.t. each edge) decide if the edge will exist w.r.t. a prob.
    threshold
    trick: PLP^T to ensure a DAG, L is lower triangular (topological order)
    P is permutation, P permute the labels
    (i,j) entry indicate j->i
    row permutation  (i,j), (k,j) becomes (k,j) (i,j)
    column permutation
    """

    def __init__(self, rng):
        self.rng = rng

    def __call__(self, num_nodes, degree):
        prob = float(degree) / (num_nodes - 1)
        # lower triagular, k=-1 is the lower off diagonal
        mat_lower_triangle_binary = np.tril(
            (self.rng.random((num_nodes, num_nodes)) < prob
             ).astype(float), k=-1
        )
        # permutes first axis only
        mat_perm = self.rng.permutation(np.eye(num_nodes, num_nodes))
        mat_b_permuted = mat_perm.T.dot(
            mat_lower_triangle_binary).dot(mat_perm)
        return mat_b_permuted



"""
test generation of true ancestral graph from ground truth DAG
"""

import numpy as np
from causalspyne.dag2ancestral import DAG2Ancestral


def test_DAG2Ancestral_path():
    """
    hide one node in a path
    """
    # lower triangular requires reverse top order
    #     0, 1, 2, 3 (column index)
    #     A, B, C, D (as parent)
    adj_matrix = np.array(
        [[0, 0, 0, 0],     # row index 0, A as child
         [1, 0, 0, 0],     # row index 1, B as child
         [0, 1, 0, 0],     # row index 2, C as child
         [0, 0, 1, 0]]     # row index 3, D as child
    )

    # - entry(1,0): entry(B, A), A->B
    # - entry(2,1): etnry(C, B), B->C
    # - entry(3,2): C->D
    #

    # if we hide B with index 1
    # 0->2 or A->C  (* new)
    # 2->3 or C->D  (edge left intact)

    # the full matrix
    #      0, 1, 2, 3 (column index, parent)
    #      A, B, C, D
    #    [[0, 0, 0, 0],   # row index 0, A as child
    #     [1, 0, 0, 0],   # row index 1, B as child
    #     [1, 1, 0, 0],   # row index 2, C as child
    #     [0, 0, 1, 0]])  # row index 3, D as child
    # new edge is (2,0) or (C, A) correponding to A->C

    # delete index 1 row
    #      0, 1, 2, 3 (column index, parent)
    #      A, B, C, D
    #    [[0, 0, 0, 0],   # row index 0, A as child
    #     [1, 1, 0, 0],   # row index 2, C as child
    #     [0, 0, 1, 0]])  # row index 3, D as child

    # delete index 1 column
    #      0, 2, 3
    #      A, C, D (parent)
    #    [[0, 0, 0],   # row index 0, A as child
    #     [1, 0, 0],   # row index 2, C as child
    #     [0, 1, 0]])  # row index 3, D as child

    sub_matrix = np.array(
        [[0, 0, 0],
         [1, 0, 0],
         [0, 1, 0]])

    # correct ancestral graph
    ancestor_graph_matrix = sub_matrix
    obj = DAG2Ancestral(adj_matrix)
    pred_ancestral_graph = obj.run([1])
    assert (ancestor_graph_matrix == pred_ancestral_graph).all()


def test_DAG2Ancestral_complicated():
    """
    hide a node with parents and multiple children, resulting in both
    kinds of directed edges as well as bidirected edges

    Original Graph:

     A -> B -> H -> {C, E, F}
     C -> D -> E
     F -> G

    """
    #        0, 1, 2, 3, 4, 5, 6, 7 (column ind, parent)
    #        A, B, H, C, D, E, F, G (as parent)
    adj_matrix = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0, 0],  # row ind 0, A as child
            [1, 0, 0, 0, 0, 0, 0, 0],  # row ind 1, B as child
            [0, 1, 0, 0, 0, 0, 0, 0],  # row ind 2, H as child
            [0, 0, 1, 0, 0, 0, 0, 0],  # row ind 3, C as child
            [0, 0, 0, 1, 0, 0, 0, 0],  # row ind 4, D as child
            [0, 0, 1, 0, 1, 0, 0, 0],  # row ind 5, E as child
            [0, 0, 1, 0, 0, 0, 0, 0],  # row ind 6, F as child
            [0, 0, 0, 0, 0, 0, 1, 0],  # row ind 7, G as child
        ]
    )

    # Updated Graph after hiding H:
    # A -> B
    # B -> {C, E, F} *new
    # C -> D -> E
    # C -> E *new
    # F -> G
    # C <-> F <-> E *new
    #         0, 1, 2, 3, 4, 5, 6, 7 (column ind, parent)
    #         A, B, H, C, D, E, F, G (as parent)

    #        [0, 0, 0, 0, 0, 0, 0, 0],  # row ind 0, A as child
    #        [1, 0, 0, 0, 0, 0, 0, 0],  # row ind 1, B as child
    #        [0, 1, 0, 0, 0, 0, 0, 0],  # row ind 2, H as child
    #        [0,(1),1, 0, 0, 0,{1},0],  # row ind 3, C as child
    #        [0, 0, 0, 1, 0, 0, 0, 0],  # row ind 4, D as child
    #        [0,(1),1,(1),1, 0,{1},0],  # row ind 5, E as child
    #        [0,(1),1,{1},0,{1},0, 0],  # row ind 6, F as child
    #        [0, 0, 0, 0, 0, 0, 1, 0],  # row ind 7, G as child

    # delete H row

    #         0, 1, 2, 3, 4, 5, 6, 7 (column ind, parent)
    #         A, B, H, C, D, E, F, G (as parent)

    #        [0, 0, 0, 0, 0, 0, 0, 0],  # row ind 0, A as child
    #        [1, 0, 0, 0, 0, 0, 0, 0],  # row ind 1, B as child
    #        [0,(1),1, 0, 0, 0,{1},0],  # row ind 3, C as child
    #        [0, 0, 0, 1, 0, 0, 0, 0],  # row ind 4, D as child
    #        [0,(1),1,(1),1, 0,{1},0],  # row ind 5, E as child
    #        [0,(1),1,{1},0,{1},0, 0],  # row ind 6, F as child
    #        [0, 0, 0, 0, 0, 0, 1, 0],  # row ind 7, G as child

    # delete H column
    #         0, 1, , 3, 4, 5, 6, 7 (column ind, parent)
    #         A, B, , C, D, E, F, G (as parent)

    #        [0, 0, , 0, 0, 0, 0, 0],  # row ind 0, A as child
    #        [1, 0, , 0, 0, 0, 0, 0],  # row ind 1, B as child
    #        [0,(1),, 0, 0, 0,{1},0],  # row ind 3, C as child
    #        [0, 0, , 1, 0, 0, 0, 0],  # row ind 4, D as child
    #        [0,(1),,(1),1, 0,{1},0],  # row ind 5, E as child
    #        [0,(1),,{1},0,{1},0, 0],  # row ind 6, F as child
    #        [0, 0, , 0, 0, 0, 1, 0],  # row ind 7, G as child

    #        submatrix
    #        A, B, C, D, E, F, G (parent)
    ancestral_graph = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0],  # A as child
            [1, 0, 0, 0, 0, 0, 0],  # B as child
            [0, 1, 0, 0, 0, 1, 0],  # C as child
            [0, 0, 1, 0, 0, 0, 0],  # D as child
            [0, 1, 1, 1, 0, 1, 0],  # E as child
            [0, 1, 1, 0, 1, 0, 0],  # F as child
            [0, 0, 0, 0, 0, 1, 0],  # G as child
        ]
    )

    obj = DAG2Ancestral(adj_matrix)
    pred_ancestral_graph = obj.run([2])  # hide H
    assert (ancestral_graph == pred_ancestral_graph).all()



"""
toplogical sort
"""

import numpy as np

from causalspyne.is_dag import is_dag


def test_is_dag0():
    """
    adj must be lower triangular
    """
    adj_matrix = np.array(
        [
            [0, 0],  #
            [1, 0],  # 1->0
        ]
    )

    is_dag(adj_matrix)


def test_is_dag():
    """
    adj must be lower triangular
    """
    adj_matrix = np.array(
        [[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0, 0, 0, 0]]  # 1->0  # 2->1  # 3->2
    )

    is_dag(adj_matrix)


def test_is_dag2():
    """
    adj must be lower triangular
    """
    adj_matrix = np.array(
        [
            # 0 1 2 3
            [0, 1, 1, 0],  # 0: 1->0, 2->0
            [0, 0, 1, 0],  # 1: 2->1
            [0, 0, 0, 1],  # 2: 3->2
            [0, 0, 0, 0],  # 3: all zero row is source
        ]
    )
    is_dag(adj_matrix)



import numpy as np
from numpy.random import default_rng
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.lines import Line2D

from causalspyne.dag_interface import MatDAG
from causalspyne.noise_idiosyncratic import Idiosyncratic
from causalspyne.data_gen import DataGen


def simpson(size_sample=200, p=0.2,
            confounder_effect: float = -5,
            treatment_effect: float = 0.1,
            propensity: float = 3,
            std: float = 1.5):
    # 0 as confounder: 0->1, 0->2, 1->2
    mat_weighted_adjacency = np.array(
        [
            # 0 1 2 3
            [0, 0, 0],  # V0: confounder, root variable
            [propensity, 0, 0],  # V1: V0->V1, propensity of getting treatment
            [confounder_effect, treatment_effect, 0],  # 2: 0->2, 1->2
        ]
    )

    dag = MatDAG(mat_weighted_adjacency,
                 name_prefix="V",
                 rng=default_rng())

    confounder = Idiosyncratic(class_name="Bernoulli",
                               dict_params={"p": p},
                               rng=default_rng()
                               )

    data_gen = DataGen(dag, edge_model=None,
                       dict_params={"std": std},
                       idiosynchratic={0: confounder})

    arr = data_gen.gen(size_sample)
    scenario = arr[:, 0]  # 1st column for scenario/confounder
    treatment = arr[:, 1]  # 2nd column: treatment
    effect = arr[:, 2]  # 3rd column: effect/performance
    return scenario, treatment, effect


def visualize_simpson(scenario, treatment, effect,
                      na_treatment="algorithm", na_confounder="scenario",
                      cut_off=0.75):
    x = treatment
    y = effect
    y = (y - np.min(y)) / (np.max(y) - np.min(y))

    ints_scenarios = np.unique(scenario)

    median_treatment = np.quantile(x, cut_off)

    discrete_treatment = np.zeros_like(x, dtype=int)  # discrete_treatment 0
    discrete_treatment[
        (scenario == ints_scenarios[0]) & (x > median_treatment)] = 1
    discrete_treatment[
        (scenario == ints_scenarios[1]) & (x > median_treatment)] = 1
    # arr_discrete_aug = np.column_stack((arr, di[crete_treatment))

    ints_treatment = np.unique(discrete_treatment)

    colors = ['green', 'orange']
    cmap = mcolors.ListedColormap(colors)
    # Boundaries separate the two values: 0 and 1
    bounds = [-0.5, 0.5, 1.5]
    norm = mcolors.BoundaryNorm(bounds, cmap.N)
    marker_map = {ints_scenarios[0]: 'o', ints_scenarios[1]: 's'}

    # fig, axs = plt.subplots(2, 2); ax=axs[0]
    fig, ax = plt.subplots()
    for ind_scenario in ints_scenarios:
        idx = np.where(scenario == ind_scenario)
        scatter = ax.scatter(
            x[idx], y[idx], c=discrete_treatment[idx],
            marker=marker_map[ind_scenario],
            edgecolor='k',
            label=f'scenario {ind_scenario}',
            cmap=cmap, norm=norm, s=100)

    ax.set_xlabel(f'jittered {na_treatment} w.r.t. {na_confounder}')
    ax.set_ylabel('performance')
    ax.tick_params(axis='x', labelbottom=False)
    ax.tick_params(axis='y', labelleft=True)
    cbar = fig.colorbar(scatter, boundaries=bounds,
                        ticks=[ints_treatment[0], ints_treatment[1]])
    cbar.ax.set_yticklabels([f'{na_treatment} {ints_treatment[0]}',
                             f'{na_treatment} {ints_treatment[1]}'])
    proxy_o = Line2D([0], [0], marker='o', color='black', linestyle='None',
                     markerfacecolor='none')
    proxy_s = Line2D([0], [0], marker='s', color='black', linestyle='None',
                     markerfacecolor='none')
    ax.legend(title=f'{na_confounder}', handles=[proxy_o, proxy_s],
              labels=[f'{na_confounder} {ints_scenarios[0]}',
                      f'{na_confounder} {ints_scenarios[1]}'])
    ax.set_title("jittered scatter plot")
    fig.savefig("simpson_jitter.pdf")

    fig, axs = plt.subplots(2, 2)
    grouped_data = [y[discrete_treatment == g] for g in ints_treatment]
    axs[0, 1].boxplot(grouped_data, labels=ints_treatment)
    axs[0, 1].set_title(f"{na_confounder}s combined")
    axs[0, 1].set_xlabel(f'{na_treatment}')
    axs[0, 1].set_ylabel('performance')
    y0 = y[scenario == ints_scenarios[0]]
    discrete_treatment0 = discrete_treatment[scenario == ints_scenarios[0]]
    grouped_data0 = [y0[discrete_treatment0 == g] for g in ints_treatment]
    axs[1, 0].boxplot(grouped_data0, labels=ints_treatment)
    axs[1, 0].set_title(f"{na_confounder} {ints_scenarios[0]}")
    axs[1, 0].set_xlabel(f'{na_treatment}')
    axs[1, 0].set_ylabel('performance')

    y1 = y[scenario == ints_scenarios[1]]
    discrete_treatment1 = discrete_treatment[scenario == ints_scenarios[1]]
    grouped_data1 = [y1[discrete_treatment1 == g] for g in ints_treatment]
    axs[1, 1].boxplot(grouped_data1, labels=ints_treatment)
    axs[1, 1].set_title(f"{na_confounder} {ints_scenarios[1]}")
    axs[1, 1].set_xlabel(f'{na_treatment}')
    axs[1, 1].set_ylabel('performance')

    fig.suptitle('simpson treatment effect')
    fig.tight_layout()
    return fig



import numpy as np
from scipy import stats


def gen_weight_matrix(rng, n=4):
    df = n + 1
    scale_matrix = np.eye(n)
    weight = stats.wishart(df=df, scale=scale_matrix, seed=rng).rvs(size=1)
    if weight.shape == ():
        weight = weight[None, None]
    return weight



"""
toplogical sort on adjacency matrix
"""

import numpy as np

from causalspyne.is_dag import is_dag


def is_binary_matrix(matrix):
    """
    check if matrix is binary
    """
    # Convert the input to a NumPy array if it isn't already
    arr = np.array(matrix)

    # Check if all elements are either 0 or 1
    return np.all((arr == 0) | (arr == 1))


def topological_sort(binary_adj_mat):
    """
    first identify all source nodes

    if binary_adj_mat[node_arrow_head][node_arrow_tail] != 0, then there is
    edge from node_arrow_tail to node_arrow_head since by default we assume
    a lower triangular matrix where the first row should be source of the graph
    since no other nodes point into it.
    """
    if not is_dag(binary_adj_mat):
        raise RuntimeError("not a DAG!")
    if not is_binary_matrix(binary_adj_mat):
        raise RuntimeError("input matrix must only have 1, 0 for counting!")
    num_nodes = len(binary_adj_mat)
    # np.sum([[0, 1], [0, 5]], axis=0)
    # array([0, 6])
    arr_node_in_degree_volatile = np.sum(binary_adj_mat, axis=1)
    # list_queue_src_node_inds initially only contains all source nodes
    list_queue_src_node_inds = [
        i for i in range(num_nodes) if arr_node_in_degree_volatile[i] == 0
    ]
    if not list_queue_src_node_inds:
        raise RuntimeError("no source nodes!")
    list_sorted_node_inds = []

    while list_queue_src_node_inds:
        node_src = list_queue_src_node_inds.pop(0)
        list_sorted_node_inds.append(node_src)
        for neighbor in range(num_nodes):
            if binary_adj_mat[neighbor][node_src] != 0:
                # arrow: node_src -> neighbor
                arr_node_in_degree_volatile[neighbor] -= 1
                if arr_node_in_degree_volatile[neighbor] == 0:
                    list_queue_src_node_inds.append(neighbor)

    if len(list_sorted_node_inds) != num_nodes:
        raise ValueError(
            "sorted nodes not completed!: "
            + str(list_sorted_node_inds)
            + str(arr_node_in_degree_volatile)
        )

    return list_sorted_node_inds



"""
test linear gaussian data gen
"""

from numpy.random import default_rng
import pandas as pd

from causalspyne.data_gen import DataGen
from causalspyne.gen_dag_2level import GenDAG2Level
from causalspyne.dag_gen import GenDAG


def test_data_gen_linear_gaussian():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=2, rng=default_rng(0)
    )
    dag = dag_gen.run()
    gen_data = DataGen(dag, rng=default_rng(0))
    arr = gen_data.gen(num_samples=200)
    arr.shape
    df = pd.DataFrame(arr, columns=dag.list_node_names)
    df.to_csv("output.csv", index=False)
    dag.to_binary_csv()



"""
toplogical sort
"""

import numpy as np

from causalspyne.utils_topological_sort import topological_sort


def test_topological_sort_linear():
    """
    adj must be lower triangular
    """
    adj_matrix = np.array(
        [[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0, 0, 0, 0]]  # 1->0  # 2->1  # 3->2
    )

    order = topological_sort(adj_matrix)
    assert order == [3, 2, 1, 0]


def test_topological_sort():
    """
    adj must be lower triangular
    """
    adj_matrix = np.array(
        [
            # 0 1 2 3
            [0, 1, 1, 0],  # 0: 1->0, 2->0
            [0, 0, 1, 0],  # 1: 2->1
            [0, 0, 0, 1],  # 2: 3->2
            [0, 0, 0, 0],  # 3: all zero row is source
        ]
    )

    order = topological_sort(adj_matrix)
    assert 3 == order[0]



"""
turn DAG into ancestral given list of variabels to hide
"""

import itertools
import copy
import numpy as np
from causalspyne.utils_closure import ancestor_matrix


def pairwise_combinations(lst):
    """
    generate pairwise combinations from a list
    """
    return itertools.combinations(lst, 2)


def to_binary(matrix):
    """
    convert a matrix to 0,1 matrix
    """
    binary_matrix = (matrix != 0).astype(int)
    return binary_matrix


class DAG2Ancestral:
    """
    turn DAG into ancestral given list of variabels to hide
    """

    def __init__(self, adj):
        self.old_adj = copy.deepcopy(adj)
        self.mat4ancestral = copy.deepcopy(self.old_adj)
        self.bmat_ancestor = None

    def pre_cal_n_hop(self):
        """
        check if one node is ancestor of another
        """
        self.bmat_ancestor = ancestor_matrix(self.old_adj)

    def run(self, list_hidden):
        """
        convert DAG to ancestral
        """
        self.pre_cal_n_hop()
        for hidden in list_hidden:
            self.deal_children(hidden)
            self.deal_parent(hidden)
        # delete first axis
        temp_mat_row = np.delete(self.mat4ancestral, list_hidden, axis=0)
        # delete second axis
        mat_adj_subgraph = np.delete(temp_mat_row, list_hidden, axis=1)
        self.mat4ancestral = mat_adj_subgraph
        return to_binary(self.mat4ancestral)

    def is_ancestor(self, global_ind_node_1, global_ind_node_2):
        """
        check if the first argument is an ancestor of the second
        """
        flag = self.bmat_ancestor[global_ind_node_2, global_ind_node_1]
        return flag

    def deal_parent(self, hidden):
        """
        connect parent of hidden and child of hidden
        """
        list_parents = self.get_list_parents(hidden)
        list_children = self.get_list_children(hidden)
        for global_ind_parent in list_parents:
            for global_ind_child in list_children:
                self.mat4ancestral[global_ind_child, global_ind_parent] = 1

    def deal_children(self, hidden):
        """
        for d_1, d_2 in children(hidden) and d_1, d_2 not connected
        """
        list_children = self.get_list_children(hidden)
        if len(list_children) < 2:
            return
        for pair in pairwise_combinations(list_children):
            c1_global_ind, c2_global_ind = pair
            if self.is_ancestor(c1_global_ind, c2_global_ind):
                # mat[i,j] means edge from j to i
                self.mat4ancestral[c2_global_ind, c1_global_ind] = 1
            elif self.is_ancestor(c2_global_ind, c1_global_ind):
                self.mat4ancestral[c1_global_ind, c2_global_ind] = 1
            else:
                self.mat4ancestral[c1_global_ind, c2_global_ind] = 1
                self.mat4ancestral[c2_global_ind, c1_global_ind] = 1

    def get_list_children(self, hidden):
        """
        adj[i,j] indicate arrow from j to i
        """
        arr = self.old_adj
        nonzero_indices = np.flatnonzero(arr[:, hidden])
        # np.nonzero() returns a tuple of arrays.
        # Each array in this tuple corresponds to a dimension of
        # the input array and contains the indices of non-zero elements
        # along that dimension.
        # nonzero_elements = arr[nonzero_indices, column_index]
        list_non_zero_indices = nonzero_indices.tolist()
        return list_non_zero_indices

    def get_list_parents(self, hidden):
        """
        adj[i,j] indicate arrow from j to i
        """
        arr = self.old_adj
        nonzero_indices = np.flatnonzero(arr[hidden, :])
        # np.nonzero() returns a tuple of arrays.
        # Each array in this tuple corresponds to a dimension of
        # the input array and contains the indices of non-zero elements
        # along that dimension.
        # nonzero_elements = arr[nonzero_indices, column_index]
        list_non_zero_indices = nonzero_indices.tolist()
        return list_non_zero_indices



from numpy.random import default_rng

from causalspyne import gen_partially_observed


def test_main():
    gen_partially_observed(
        degree=2,
        list_confounder2hide=[0.5, 0.9],
        size_micro_node_dag=4,
        num_macro_nodes=4,
        num_sample=200,
        rng=default_rng(0),
    )



import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
from causalspyne.implicit_gen_Sigma import gen_spectrum


# Generate sample data

# samples = np.random.normal(size=10000)
samples = [gen_spectrum() for _ in range(10000)]
# Create histogram
fig, ax = plt.subplots(figsize=(8, 6))

hist, bins, _ = ax.hist(samples, bins=30, density=True, alpha=0.7,
                        label="Histogram")

# Calculate bin centers for PDF plotting
bin_centers = 0.5 * (bins[1:] + bins[:-1])

# Compute PDF
pdf = stats.norm.pdf(bin_centers)

# Plot PDF
ax.plot(bin_centers, pdf, 'r-', label="PDF")

# Customize plot
ax.set_xlabel('eigenvalue')
ax.set_ylabel('Density')
ax.set_title('Histogram of maximum eigenvalue distribution')
ax.legend()

plt.show()



import numpy as np

def estimate_spectrum_radius(arr_df):
    mat_corr = np.corrcoef(arr_df.T)
    mat_cov = np.cov(arr_df.T)
    mat_precision = np.linalg.inv(mat_cov)
    arr_eig_values, eigvec = np.linalg.eig(mat_precision)
    return np.max(arr_eig_values)



"""
concrete class to generate simple DAGs

"""

import warnings
from numpy.random import default_rng

from causalspyne.erdo_renyi_plp import Erdos_Renyi_PLP
from causalspyne.dag_interface import MatDAG
from causalspyne.weight import WeightGenWishart
from causalspyne.dag_manipulator import DAGManipulator


class GenDAG:
    def __init__(self, num_nodes, degree, obj_gen_weight=None,
                 rng=default_rng()):
        """
        degree: expected degree for each node
        """
        self.num_nodes = num_nodes
        self.degree = degree
        self.strategy_gen_dag = Erdos_Renyi_PLP(rng)
        self.obj_gen_weight = obj_gen_weight
        if obj_gen_weight is None:
            self.obj_gen_weight = WeightGenWishart(rng=rng)
        self.dag_manipulator = None
        self.rng = rng

    def gen_dag(self, num_nodes=None, prefix="", target_num_confounder=2):
        """
        generate DAG and wrap it around with interface
        """
        if num_nodes is None:
            num_nodes = self.num_nodes
        mat_skeleton = self.strategy_gen_dag(num_nodes, self.degree)

        mat_mask = (mat_skeleton != 0).astype(float)
        mat_weight = self.obj_gen_weight.gen(num_nodes)
        # Hardarmard product
        mat_weighted_adjacency = mat_mask * mat_weight

        dag = MatDAG(mat_weighted_adjacency, name_prefix=prefix, rng=self.rng)
        self.dag_manipulator = DAGManipulator(dag,
                                              self.obj_gen_weight, self.rng)
        ind_arbitrary = dag.get_top_last()
        counter = 0
        for _ in range(dag.num_nodes):
            flag_success = self.dag_manipulator.mk_confound(
                ind_arbitrary_confound_input=ind_arbitrary)
            if not flag_success:
                counter += 1
            if dag.num_confounder >= target_num_confounder:
                break
            # FIXME: it can be the new ind_arbitrary has been tried out already
            ind_arbitrary = dag.climb(ind_arbitrary)
            if ind_arbitrary is None:
                break
        num_confounder = len(dag.list_confounder)
        if num_confounder < target_num_confounder and \
                dag.num_nodes - target_num_confounder > 1:
            warnings.warn(
                f"\n failed to ensure {target_num_confounder} confounders for \
                adjacency matrix \n{dag.mat_adjacency}, \
                \n after {counter} failed trials, \
                \n{num_confounder} confounders only")
        return dag



from causalspyne import gen_partially_observed
from causalspyne.utils_cov_precision import estimate_spectrum_radius


subview = gen_partially_observed(size_micro_node_dag=3,
                                 num_macro_nodes=3,
                                 degree=2,
                                 list_confounder2hide=[0],
                                 num_sample=200,
                                 graphviz=True)

radius = estimate_spectrum_radius(subview.data)
print(radius)



"""
test data and DAG subview
"""
import numpy as np
from numpy.random import default_rng
from causalspyne.gen_dag_2level import GenDAG2Level
from causalspyne.dag_gen import GenDAG
from causalspyne.dag_viewer import DAGView
from causalspyne.dag2ancestral import DAG2Ancestral



def test_data_dag2ancestral():
    """
    """
    simple_dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=3, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete")

    obj = DAG2Ancestral(dag.mat_adjacency)
    pred_ancestral_graph = obj.run([1, 2])
    print("\n")
    print(pred_ancestral_graph)



"""
generate Linear Gaussian
"""

import numpy as np
from numpy.random import default_rng

from causalspyne.noise_idiosyncratic import Idiosyncratic
from causalspyne.edge_models import EdgeModelLinear


class DataGen:
    def __init__(self, dag, edge_model=None,
                 dft_noise: str="Gaussian",
                 dict_params: dict={},
                 idiosynchratic: dict[int, Idiosyncratic]={},
                 rng=default_rng(0)):
        self.dag = dag
        self.dict_idiosyncratic = idiosynchratic

        self.idiosyncratic = Idiosyncratic(class_name=dft_noise,
                                           rng=rng,
                                           dict_params=dict_params)
        self.edge_model = edge_model
        if edge_model is None:
            self.edge_model = EdgeModelLinear(self.dag)

    def gen(self, num_samples):
        """
        Generate linear Gaussian data from a given DAG.

        Parameters:
        - num_samples: int, number of samples to generate.
        - noise_std: float, standard deviation of the Gaussian noise.

        Returns:
        - data: np.ndarray, generated data of shape (num_samples, num_nodes).
        """
        list_ind_nodes_topo_order = self.dag.topological_sort()

        # Number of nodes
        num_nodes = len(list_ind_nodes_topo_order)

        # Initialize the data matrix
        data = np.zeros((num_samples, num_nodes))

        # Generate data for each node in topological order
        for node in list_ind_nodes_topo_order:
            list_parents_inds = self.dag.get_list_parents_inds(node)
            if node in self.dict_idiosyncratic:
                noise = self.dict_idiosyncratic[node].gen(num_samples)
            else:
                noise = self.idiosyncratic.gen(num_samples)
            data[:, node] = noise
            if list_parents_inds:
                bias = self.edge_model.run(node, data[:, list_parents_inds])
                data[:, node] += bias
        return data



from causalspyne.poset_approach import partitions


def test_lattice_enumerate():
    gen = partitions(2)
    # gen = partitions(5)
    # gen = partitions(10)
    # gen = partitions(15)
    len(list(gen))



"""
check if transitive closure correct for ancestor relationship determination
"""
import numpy as np
from causalspyne.utils_closure import ancestor_matrix


def test_utils_transitive_closure():
    """
    check if transitive closure correct
    """
    # Example usage note this adj matrix is opposite of our convention
    # A->B, B->C, C->D
    #     A, B, C, D
    adj_matrix = np.array(
        [[0, 1, 0, 0],     # A row as parent
         [0, 0, 1, 0],     # B row as parent
         [0, 0, 0, 1],     # C row as parent
         [0, 0, 0, 0]])    # D row as parent
    pred_ancestor_mat = ancestor_matrix(adj_matrix)
    print("pred Ancestor matrix:")
    print(pred_ancestor_mat.astype(int))
    # A->C(new)
    # A->D(new)
    # B->D(new)
    #  A B C D
    # [[
    #  0 1 1 1]   # A row as parent
    # [0 0 1 1]   # B row as parent
    # [0 0 0 1]   # C row as parent
    # [0 0 0 0]]  # D row as parent

    #     A, B, C, D
    ground_truth_ancestor_matrix = np.array(
        [[0, 1, 1, 1],     # A row as parent
         [0, 0, 1, 1],     # B row as parent
         [0, 0, 0, 1],     # C row as parent
         [0, 0, 0, 0]])    # D row as parent
    assert (ground_truth_ancestor_matrix == pred_ancestor_mat).all()


def test_utils_transitive_closure_transpose():
    """
    check if transitive closure correct
    """
    # Example usage note this adj matrix is opposite of our convention
    # A->B, B->C, C->D
    #     A, B, C, D (column index as parent)
    adj_matrix = np.array(
        [[0, 0, 0, 0],     # A row as child
         [1, 0, 0, 0],     # B row as child
         [0, 1, 0, 0],     # C row as child
         [0, 0, 1, 0]])    # D row as child
    pred_ancestor_mat = ancestor_matrix(adj_matrix)
    print("pred Ancestor matrix:")
    print(pred_ancestor_mat.astype(int))
    # A->C(new)
    # A->D(new)
    # B->D(new)
    #  A   B   C D
    # [[
    #  0   0   0 0]   # A row as child
    # [1   0   0 0]   # B row as child
    # [(1) 1   0 0]   # C row as child
    # [(1) (1) 1 0]]  # D row as child

    #     A, B, C, D
    ground_truth_ancestor_matrix = np.array(
        [[0, 0, 0, 0],     # A row as parent
         [1, 0, 0, 0],     # B row as parent
         [1, 1, 0, 0],     # C row as parent
         [1, 1, 1, 0]])    # D row as parent
    assert (ground_truth_ancestor_matrix == pred_ancestor_mat).all()


def test_utils_transitive_closure_transpose2():
    """
    check if transitive closure correct
    """
    # Example usage note this adj matrix is opposite of our convention
    # A->B, B->C, C->D, A->c
    #     A, B, C, D (column index as parent)
    adj_matrix = np.array(
        [[0, 0, 0, 0],     # A row as child
         [1, 0, 0, 0],     # B row as child
         [1, 1, 0, 0],     # C row as child
         [0, 0, 1, 0]])    # D row as child
    pred_ancestor_mat = ancestor_matrix(adj_matrix)
    print("pred Ancestor matrix:")
    print(pred_ancestor_mat.astype(int))
    # A->D(new)
    # B->D(new)
    #  A   B   C D
    # [[
    #  0   0   0 0]   # A row as child
    # [1   0   0 0]   # B row as child
    # [(1) 1   0 0]   # C row as child
    # [(1) (1) 1 0]]  # D row as child

    #     A, B, C, D
    ground_truth_ancestor_matrix = np.array(
        [[0, 0, 0, 0],     # A row as parent
         [1, 0, 0, 0],     # B row as parent
         [1, 1, 0, 0],     # C row as parent
         [1, 1, 1, 0]])    # D row as parent
    assert (ground_truth_ancestor_matrix == pred_ancestor_mat).all()



"""
"""

from causallearn.utils.DAG2PAG import dag2pag
from causallearn.graph.Dag import Dag
import numpy as np

# class ancestral_shd():
# def __init__(mat_gt_ancestral, mat_fci):
#    self.mat_gt_ancestral = mat_gt_ancestral


import numpy as np


def structural_hamming_distance(true_dag, true_hidden_nodes, prediction):
    """
    Compute the standardized structural Hamming distance between two graphs.

    Parameters:
    true_dag (numpy.ndarray): Target graph adjacency matrix
    true_hidden_nodes (list): list of nodes in DAG to hide
    prediction (numpy.ndarray): Predicted graph adjacency matrix  (pag.graph
    from causal-learn)
    double_for_anticausal (bool): Count badly oriented edges as two mistakes

    Returns:
    int: Structural Hamming Distance
    """
    n = len(true_dag) - len(true_hidden_nodes)
    if (n, n) != prediction.shape:
        raise ValueError("Graphs must have the same number of nodes")

    cl_dag = Dag(range(len(true_dag)))
    for ch, pa in np.argwhere(true_dag):
        cl_dag.add_directed_edge(pa, ch)
    true_pag = dag2pag(cl_dag, true_hidden_nodes)

    total_shd = np.sum(true_pag.graph != prediction)

    return total_shd / (n**2 - n)



"""
draw DAG using networkx
"""

import matplotlib.pyplot as plt
import networkx as nx
try:
    import pygraphviz
    has_graphviz = True
    from networkx.drawing.nx_agraph import graphviz_layout
except ImportError:
    graphviz_layout = None
    has_graphviz = False



def draw_dags_nx(
    adj_matrix, dict_ind2name=None, title="dag", ax=None, show=False,
    graphviz=False
):
    """
    networkx adjacency matrix (i,j) entry refers to edge from i pointing to j,
    which is opposite to the CausalSpyne convention
    """
    graphviz = has_graphviz and graphviz
    plt.close("all")
    nx_graph = nx.from_numpy_array(adj_matrix.transpose(),
                                   create_using=nx.DiGraph)
    if dict_ind2name:
        nx_graph = nx.relabel_nodes(nx_graph, dict_ind2name)
    if graphviz:
        pos = graphviz_layout(nx_graph, prog="dot")
    else:
        pos = nx.spring_layout(nx_graph, k=0.5, scale=2)
    nx.draw(nx_graph, pos=pos, ax=ax, arrows=True, with_labels=True,
            node_color="white")
    if ax is None:
        plt.title(title)
        plt.axis("off")
        plt.savefig(title + ".pdf", format="pdf")
        plt.savefig(title + ".svg", format="svg")
    if show:
        plt.show()



from numpy.random import default_rng

from causalspyne.wishart import gen_weight_matrix


class WeightGenUniform:
    def __init__(self, list_weight_range, prob_neg_weights=0.5, rng=default_rng(0)):
        self.list_weight_range = list_weight_range
        self.prob_neg_weights = prob_neg_weights
        self.rng = rng

    def gen(self, num_nodes):
        """
        generate complete graph, fully connected
        """
        mat_weight = self.rng.uniform(
            low=self.list_weight_range[0],
            high=self.list_weight_range[1],
            size=[num_nodes, num_nodes],
        )

        # set some edges randomly to negative: e.g. x_i = 2x_j - 3x_k
        random_mask = self.rng.choice(
            [True, False],
            (num_nodes, num_nodes),
            (self.prob_neg_weights, 1 - self.prob_neg_weights),
        )
        mat_weight[random_mask] *= -1
        if mat_weight.size == 1:
            return mat_weight.item()
        return mat_weight


class WeightGenWishart(WeightGenUniform):
    def __init__(self, prob_neg_weights=0.5, rng=default_rng(0)):
        self.prob_neg_weights = prob_neg_weights
        self.rng = rng

    def gen(self, num_nodes):
        """
        generate complete graph, fully connected
        """
        mat_weight = gen_weight_matrix(self.rng, num_nodes)

        # set some edges randomly to negative: e.g. x_i = 2x_j - 3x_k
        random_mask = self.rng.choice(
            [True, False],
            (num_nodes, num_nodes),
            (self.prob_neg_weights, 1 - self.prob_neg_weights),
        )
        mat_weight[random_mask] *= -1
        if mat_weight.size == 1:
            return mat_weight.item()
        return mat_weight



"""
test if shd runs
"""

import numpy as np
from causalspyne.ancestral_shd import structural_hamming_distance


def test_shd():
    """
    test shd runs
    """
    target_graph = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

    predicted_graph = np.array([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    shd = structural_hamming_distance(target_graph, [], predicted_graph)
    print(f"Structural Hamming Distance: {shd}")



"""
test if ancestral acc runs
"""

import numpy as np
from causalspyne.ancestral_acc import ancestral_acc
from causalspyne import gen_partially_observed


def test_ancestral_acc():
    """
    test if ancestral accuracy runs
    """
    subview = gen_partially_observed(
        size_micro_node_dag=3,
        num_macro_nodes=2,
        degree=2,  # average vertex/node degree
        list_confounder2hide=[
            0.5,
            1.0,
        ],  # choie of confounder to hide: percentile or index of all toplogically sorted confounders
        num_sample=200,
        rng=np.random.default_rng(1),
        graphviz=False
    )
    pred_order_inds = \
        [subview.dag._dict_node_names2ind[name]
         for name in subview.dag.list_node_names]
    acc = ancestral_acc(true_dag=subview.dag,
                        pred_order=pred_order_inds)

    print(f"ancestral acc: {acc}")



import sys
import numpy as np


class Idiosyncratic:
    def __init__(self, class_name, rng, dict_params={}):
        """ """
        class_gen = getattr(sys.modules[__name__], class_name)
        if bool(dict_params):
            self.noise = class_gen(rng=rng,
                                   dict_params=dict_params)
        else:
            self.noise = class_gen(rng=rng)

    def gen(self, num_samples):
        noise = self.noise.gen(num_samples)
        return noise


class Gaussian:
    def __init__(self, rng, dict_params={"std":1.0}):
        self.mean = 0.0
        self.noise_std = dict_params["std"]
        self.rng = rng

    def gen(self, num_samples):
        noise = self.rng.normal(loc=self.mean, scale=self.noise_std,
                                size=num_samples)
        return noise


class Gamma:
    def __init__(self, rng, dict_params={"shape":1, "scale":2.0}):
        # shape: The shape parameter of the gamma distribution (k > 0)
        # scale: The scale parameter of the gamma distribution
        # (theta > 0, default is 1.0)
        self.par_shape = dict_params["shape"]
        self.par_scale = dict_params["scale"]
        self.rng = rng

    def gen(self, num_samples):
        """
        size: The number of samples to generate (optional)
        """
        noise = self.rng.gamma(shape=self.par_shape,
                               scale=self.par_scale,
                               size=num_samples)
        return noise


class Bernoulli:
    def __init__(self, rng, dict_params={"p":0.5}):
        self.p = dict_params["p"]
        self.rng = rng

    def gen(self, num_samples):
        noise = self.rng.binomial(size=num_samples, p=self.p, n=1)
        return noise



import os
from contextlib import contextmanager


@contextmanager
def chdir(path):
    original_dir = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(original_dir)



"""
insert edges to have more confounders
"""
import numpy as np


class DAGManipulator:
    """
    insert edges to have more confounders
    """
    def __init__(self, dag, obj_gen_weight, rng):
        self.dag = dag
        self._obj_gen_weight = obj_gen_weight
        self.rng = rng

    def mk_confound(self, ind_arbitrary_confound_input=None,
                    continue_top_up=True,
                    skip_sink=False):
        """
        ind_arbitrary_confound_input is the arbitrary index for a node, if
        does not set as argument, a random index will be generated, the
        function make the current vertex confounder
        continue_top_up: if not successful, move up toplogically to try again
        skip_sink: if a node is sink node (can be a high toplogical order),
        then does not add edge to other nodes with lower toplogical order
        """
        ind_arbitrary_confound = ind_arbitrary_confound_input
        if ind_arbitrary_confound is None:
            ind_arbitrary_confound = \
                self.rng.choice(self.dag.list_ind_nodes_sorted)
            # list_ind_nodes_sorted looks like [8, 3, 10, 9, 100],
            # each entry is an arbitrary index/name for the node
        # rembmer (i,j) entry means there is arrow from j to i
        # count how many children this current node has
        nnzero = np.count_nonzero(
            self.dag.mat_adjacency[:, ind_arbitrary_confound])
        flag_success = False
        if nnzero == 0:  # 0 means sink node
            # sink node can also be quite high in toplogical rank
            if not skip_sink:
                flag_success_1 = self.add_new_edge(ind_arbitrary_confound)
                flag_success_2 = self.add_new_edge(ind_arbitrary_confound)
                flag_success = flag_success_1 & flag_success_2
        elif nnzero == 1:  # not a sink node but only parent a single child
            # add another child to this current node
            flag_success = self.add_new_edge(ind_arbitrary_confound)
            # not successul in making this node to have two children
        else:  # already a confounder
            flag_success = False  # only when new confounder
        if flag_success:
            return True
        # now flag_success must be False
        if continue_top_up:
            # increase the toplogical order and try until success
            pos = self.dag.global_arbitrary_ind2topind(ind_arbitrary_confound)
            if pos - 1 < 0:
                return False
            ind_arbitrary = self.dag.top_ind2global_arbitrary(pos - 1)
            return self.mk_confound(
                ind_arbitrary_confound_input=ind_arbitrary)
        return False

    def add_new_edge(self, ind_arbitrary_confound):
        """
        insert a new edge pointing to lower toplogical order
        """
        # get toplogical order of this node
        pos = self.dag.global_arbitrary_ind2topind(ind_arbitrary_confound)
        # randomly choose one node which ranked later than the
        # current node
        if pos + 1 == self.dag.num_nodes:
            return False
        ind_toplogical_arrow_head = self.rng.integers(
            pos + 1, self.dag.num_nodes)
        ind_arbitrary_arrow_head = self.dag.top_ind2global_arbitrary(
            ind_toplogical_arrow_head)

        if (
            self.dag.mat_adjacency[
                ind_arbitrary_arrow_head, ind_arbitrary_confound]
            == 0
        ):
            self.dag.mat_adjacency[
                ind_arbitrary_arrow_head, ind_arbitrary_confound
            ] = self._obj_gen_weight.gen(1)
            self.dag.check()  # check if still a DAG
            return True
        # in very rare cases, the randomly chosen node is already a
        # child of this current node
        return False



"""
"""
import numpy as np
from numpy.random import default_rng
from causalspyne.erdo_renyi_plp import Erdos_Renyi_PLP
from causalspyne.implicit_omega import gen_omega, get_extreme_eigenvalue


def get_max_degree(mat_w):
    d_max = 0
    for ind_row, row in enumerate(mat_w):
        # w_{i,j} j->i
        in_degree = len(np.nonzero(row))
        out_degree = len(np.nonzero(mat_w[:, ind_row]))
        if in_degree + out_degree > d_max:
            d_max = in_degree + out_degree
    return d_max


def gen_joint_w_omega(p_ob_v=4, ave_degree=2, max_omega=0.7):
    mat_w = Erdos_Renyi_PLP(default_rng(0))(num_nodes=p_ob_v,
                                            degree=ave_degree)
    print(f"adjacency: \n {mat_w}")
    mat_omega, rho = gen_omega(p_ob_v, max_omega)
    print(f"omega: \n {mat_omega}")
    max_eig, min_eig = get_extreme_eigenvalue(mat_omega)
    print(f"min, max eigenvalue {min_eig, max_eig}")
    print(f"rho, 2*rho: {rho}, {2 * rho}")
    return mat_w, mat_omega


if __name__ == "__main__":
    gen_joint_w_omega()



"""
ancestral accuracy calculate the percentage of correct prediction
among all pairwise variables' ancestral relationships (binary)
"""
from itertools import combinations
import warnings

from causalspyne.dag2ancestral import DAG2Ancestral


def ancestral_acc(true_dag, pred_order, list_hidden_nodes=None):
    """
    Parameters:
    true_dag: Target causalspyne graph
    pred_order:
    ISSUE #34, the pairwise relationship can be + or - ancestral, but can also be no ancestral relationship
    list_hidden_nodes (list): list of nodes in DAG to hide, here we use pairwise combinations of pred_order, which
    already assumes the algorithm has a clear + order only. 
    Returns:
    float:
    """
    dag2ancestral = DAG2Ancestral(true_dag.mat_adjacency)
    dag2ancestral.pre_cal_n_hop()
    size_dag = true_dag.num_nodes
    if list_hidden_nodes is not None:
        n_vars = size_dag - len(list_hidden_nodes)
    else:
        n_vars = size_dag
    if n_vars != len(pred_order):
        warnings.warn(f"predicted causal order {pred_order} does not \
                      have the same number of observables!, hidden are: {list_hidden_nodes} \
                      now forcing number of variables to be the lengh of pred_order")
        n_vars = len(pred_order)

    pairwise_combinations = list(combinations(pred_order, 2))
 
    n_correct = 0
    for pair in pairwise_combinations:
        ancestor, offspring = pair
        if dag2ancestral.is_ancestor(ancestor, offspring):
            n_correct += 1
    num_combos = n_vars * (n_vars - 1) /2
    return float(n_correct) / num_combos



"""
DFS to check if is DAG, use python inner function to access global variable
"""


def is_dag(adjacency_matrix):
    """
    adjaceny_matrx[u][v] \\neq 0 \\iff u <- v
    """
    n = len(adjacency_matrix)
    set_visited = set()
    set_path = set()  # A->[B]->[C]->A

    def dfs(ind_node):
        set_visited.add(ind_node)
        set_path.add(ind_node)

        for arrow_head in range(n):
            if adjacency_matrix[arrow_head][ind_node] != 0:  # arrow head
                if arrow_head in set_path:  # # A->B->C->A
                    return False  # not a DAG, bottom level return
                if arrow_head not in set_visited:
                    if not dfs(arrow_head):  # recursion
                        return False  # recursive return
        # exit of recursion: if arrow_head is visited
        set_path.remove(ind_node)
        return True

    for ind_node in range(n):
        if ind_node not in set_visited:
            if not dfs(ind_node):
                return False

    return True



"""
A graph is simpler than simplicial complex in that it only characterize
pairwise relationship, which enables us to project a complicated graph in
causallearn to ancestral ADMG (i.e. ancestral graph with only directed edges)

We of course lose information via projection.

This module offers another function to extract causal order from
causal-learn
"""
from causallearn.graph.NodeType import NodeType


def project_causallearn_g2ancestral_admg(mat_graph_causallearn):
    """
    A general graph in causal-learn is not necessarily a PAG (it does have
    an attribute to denote whether it is a PAG or not)
    When it is a PAG, like the return from FCI:
    https://causal-learn.readthedocs.io/en/latest/search_methods_index/Constraint-based%20causal%20discovery%20methods/FCI.html#usage
    g.graph is a PAG
    A->B:   G[A, B]=-1, G[B, A]=1, to our format, G[A, B]=0, G[B, A]=1
    A.->B:  G[A, B]=2, G[B, A]=1, B is not an ancestor of A,
    to our format, G[A, B]=0, G[B, A]=1
    A.-.B:  G[A, B]=2, G[B,A]=2, no set d-separates A and B, ?????
    A<->B: there is common latent cause, G[A, B]=1, G[B, A]=1,
    no need to convert
    """
    for i in range(len(mat_graph_causallearn)):
        for j in range(len(mat_graph_causallearn[i])):
            if mat_graph_causallearn[i][j] != 1:
                mat_graph_causallearn[i][j] = 0
    return mat_graph_causallearn


def get_causalearn_order(g_causal_learn, node_names=None):
    """
    get causal order from causal learn graph using api like
    'node.get_node_type() == NodeType.LATENT:'
    """
    real_name_order = []
    real_na_order_latent = []
    nodes_order = g_causal_learn.get_causal_ordering()
    for node in nodes_order:
        if node.get_node_type() == NodeType.LATENT:
            # GIN-LINLAM
            real_na_order_latent.append(node.get_node_type)
            continue
        fake_name = node.get_name()
        index = int(fake_name.removeprefix("X")) - 1
        real_name = node_names[index]
        print(f"{real_name}")
        real_name_order.append(real_name)
        real_na_order_latent.append(real_name)
    print(f"{real_name_order}")
    return real_name_order, real_na_order_latent



"""
2-level DAG generation
"""

from numpy.random import default_rng

from causalspyne.dag_stack_indexer import DAGStackIndexer
from causalspyne.dag_manipulator import DAGManipulator
from causalspyne.weight import WeightGenWishart


class GenDAG2Level:
    """
    generate a DAG with 2 levels: first level generate macro nodes, second
    level populate each macro node
    """

    def __init__(
        self, dag_generator, num_macro_nodes, max_num_local_nodes=4, rng=default_rng()
    ):
        self.dag_generator = dag_generator
        self.num_macro_nodes = num_macro_nodes
        self.max_num_local_nodes = max_num_local_nodes

        self.global_dag_indexer = None
        self.dag_backbone = None
        self.dict_macro_node2dag = {}
        self.dag_refined = None
        self.rng = rng
        self.dag_manipulator = None

    def populate_macro_node(self):
        """
        replace a macro node into a DAG
        """
        # iterate each macro node
        for name in self.dag_backbone.list_node_names:
            num_nodes = self.rng.integers(2, self.max_num_local_nodes + 1)
            self.dict_macro_node2dag[name] = self.dag_generator.gen_dag(
                num_nodes=num_nodes, prefix=name
            )
        self.global_dag_indexer = DAGStackIndexer(self)

    def interconnection(self):
        """
        connect macro nodes with edges
        """
        # iterate over the Macro-DAG edges
        for arc in self.dag_backbone.list_arcs:
            self.connect_macro_node_via_local_node(arc)

    def connect_macro_node_via_local_node(self, arc):
        """
        connect macro-DAG node edge (i,j) via local nodes
        """
        macro_arrow_tail, macro_arrow_head = arc
        _, ind_local_tail = self.dict_macro_node2dag[macro_arrow_tail].sample_node()
        _, ind_local_head = self.dict_macro_node2dag[macro_arrow_head].sample_node()

        ind_macro_tail = self.dag_backbone.get_node_ind(macro_arrow_tail)
        ind_macro_head = self.dag_backbone.get_node_ind(macro_arrow_head)

        ind_global_tail = self.global_dag_indexer.get_global_ind(
            ind_macro_tail, ind_local_tail
        )
        ind_global_head = self.global_dag_indexer.get_global_ind(
            ind_macro_head, ind_local_head
        )

        self.dag_refined.add_arc_ind(ind_global_tail, ind_global_head)

    def inject_additional_confounder(self):
        """
        make confounder in the big graph
        """
        obj_gen_weight = WeightGenWishart(rng=self.rng)
        self.dag_manipulator = DAGManipulator(self.dag_refined,
                                              obj_gen_weight, self.rng)

        ind_arbitrary = self.dag_refined.get_top_last()
        self.dag_manipulator.mk_confound(ind_arbitrary)
        print(self.dag_refined.num_confounder)
        ind_arbitrary = self.dag_refined.climb(ind_arbitrary)
        self.dag_manipulator.mk_confound(ind_arbitrary)
        print(self.dag_refined.num_confounder)

    def run(self):
        """
        generation
        """

        # generate dag_backbone DAG with only macro nodes
        self.dag_backbone = self.dag_generator.gen_dag(self.num_macro_nodes)
        self.populate_macro_node()
        self.interconnection()
        self.dag_refined.check()
        self.inject_additional_confounder()
        return self.dag_refined



"""
index node globally when stacking several dags together
"""

from scipy.linalg import block_diag
from causalspyne.dag_interface import MatDAG


class DAGStackIndexer:
    """
    stack DAG indexer
    """

    def __init__(self, host):
        """ """
        self.host = host
        self.host.dag_refined = self.stack_dags()

    def get_global_ind(self, ind_macro_node, ind_local_node):
        """
        get the global index of a local node
        """
        return self.list_accum_count[ind_macro_node] + ind_local_node

    def stack_dags(self):
        """
        stack dictionary of DAG into a block diagnoal matrix
        """
        mat_stacked_dag = block_diag(
            *(dag.mat_adjacency for dag in self.host.dict_macro_node2dag.values())
        )
        dag_stacked = MatDAG(mat_stacked_dag, rng=self.host.rng)
        self.dict_num = {
            key: dag.num_nodes for key, dag in self.host.dict_macro_node2dag.items()
        }
        list_accum_count = list(fun_accum_sum(self.dict_num))
        self.list_accum_count = [0] + list_accum_count[:-1]
        dag_stacked.gen_node_names_stacked(self.host.dict_macro_node2dag)
        return dag_stacked


def fun_accum_sum(dict_num):
    """
    count accumulative sum of a dictionary
    e.g. dict_dag = {'0':3, '1':5, '2':4}
    """
    count = 0
    for num in dict_num.values():
        count += num
        yield count



import matplotlib.pyplot as plt
from causalspyne.dataset import simpson, visualize_simpson


scenario, treatment, effect = simpson(200)
plot = visualize_simpson(scenario, treatment, effect)
plot.savefig('simpson_scatter_plot.pdf', format='pdf')
plt.show()




"""
test data and DAG subview
"""

from numpy.random import default_rng

from causalspyne.gen_dag_2level import GenDAG2Level
from causalspyne.dag_gen import GenDAG
from causalspyne.dag_viewer import DAGView


def test_data_dag_subview():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=2, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete")
    subview = DAGView(dag=dag, dft_noise="Gaussian")
    # subview.run(num_samples=200)
    subview.run(num_samples=200, list_nodes2hide=[1, 2])
    subview.to_csv()
    subview.visualize(title="dag_marginal")


def test_data_dag_subview_confounder():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=3, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=2, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete_confounder")

    subview = DAGView(dag=dag,dft_noise="Gaussian")
    subview.run(num_samples=200, confound=True, list_nodes2hide=[0])
    subview.to_csv()
    subview.visualize(title="dag_marginal_confounder")


def test_data_dag_subview_confounder_percentage():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=4, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=4, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete_confounder")

    subview = DAGView(dag=dag,dft_noise="Gaussian")
    subview.run(num_samples=200, confound=True, list_nodes2hide=[0.99])
    subview.to_csv()
    subview.visualize(title="dag_marginal_confounder")


def test_data_dag_subview_confounder_percentage():
    """
    test linear gaussian data gen
    """
    simple_dag_gen = GenDAG(num_nodes=4, degree=2, rng=default_rng(0))
    dag_gen = GenDAG2Level(
        dag_generator=simple_dag_gen, num_macro_nodes=4, rng=default_rng(0)
    )
    dag = dag_gen.run()
    dag.visualize(title="dag_complete_confounder2")

    subview = DAGView(dag=dag,dft_noise="Gaussian")
    subview.run(num_samples=200, confound=True, list_nodes2hide=[0.1, 1.0])
    subview.to_csv()
    subview.visualize(title="dag_marginal_confounder2")
